#!/usr/bin/env python

import yaml


class MachineModel:
    def __init__(self, path_to_yaml):
        self._path = path_to_yaml
        self._data = {}
        with open(path_to_yaml, 'r') as f:
            self._data = yaml.load(f)

        self.__getitem__ = self._data.__getitem__

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self._path))



#!/usr/bin/env python

import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot as plt
import yaml
from pprint import pprint

from prefixedunit import PrefixedUnit


def frange(start, stop, step=1.0):
    f = start
    while f < stop:
        f += step
        yield f

# Input (usually from ECM model)
result = {
    'min performance': 11175000000.0, 'bottleneck level': 2,
    'mem bottlenecks': [{'performance': PrefixedUnit(24474545454.545452, '', 'FLOP/s'),
                         'bandwidth': PrefixedUnit(89.74, u'G', u'B/s'),
                         'arithmetic intensity': 0.2727272727272727,
                         'bw kernel': 'triad', 'level': 'L1-L2'},
                        {'performance': PrefixedUnit(12957000000.0, '',
                         'FLOP/s'), 'bandwidth': PrefixedUnit(43.19, u'G', u'B/s'),
                         'arithmetic intensity': 0.3, 'bw kernel': 'triad', 'level': 'L2-L3'},
                        {'performance': PrefixedUnit(11175000000.0, '', 'FLOP/s'),
                         'bandwidth': PrefixedUnit(22.35, u'G', u'B/s'),
                         'arithmetic intensity': 0.5, 'bw kernel': 'triad', 'level': 'L3-MEM'}]}
machine = yaml.load(open('machine-files/emmy.yaml'))
max_flops = machine['clock']*sum(machine['FLOPs per cycle']['DP'].values())
max_flops.unit = "FLOP/s"

pprint(result)
pprint(max_flops)

# Plot configuration
height = 0.8

fig = plt.figure(frameon=False)
ax = fig.add_subplot(1, 1, 1)

yticks_labels = []
yticks = []
xticks_labels = []
xticks = [2.**i for i in range(-4, 4)]

ax.set_xlabel('arithmetic intensity [FLOP/byte]')
ax.set_ylabel('performance [FLOP/s]')

# Upper bound
x = list(frange(min(xticks), max(xticks), 0.01))
bw = float(result['mem bottlenecks'][result['bottleneck level']]['bandwidth'])
ax.plot(x, map(lambda x: min(bw*x, float(max_flops)), x))

# Code location
perf = min(
    float(max_flops),
    float(result['mem bottlenecks'][result['bottleneck level']]['performance']))
arith_intensity = result['mem bottlenecks'][result['bottleneck level']]['arithmetic intensity']
ax.plot(arith_intensity, perf, 'r+', markersize=12, markeredgewidth=4)

#ax.tick_params(axis='y', which='both', left='off', right='off')
#ax.tick_params(axis='x', which='both', top='off')
ax.set_xscale('log', basex=2)
ax.set_yscale('log')
ax.set_xlim(min(xticks), max(xticks))
#ax.set_yticks([perf, float(max_flops)])
ax.set_xticks(xticks+[arith_intensity])
ax.grid(axis='x', alpha=0.7, linestyle='--')
#fig.savefig('out.pdf')
plt.show()



#!/usr/bin/env python
from __future__ import print_function

import yaml
import re


class PrefixedUnit(yaml.YAMLObject):
    PREFIXES = {'k': 1e3, 'M': 1e6, 'G': 1e9, 'T': 1e13, 'P': 1e16, 'E': 1e19, 'Z': 1e21, 'Y': 1e24,
                '': 1}

    yaml_loader = yaml.Loader
    yaml_dumper = yaml.Dumper

    yaml_tag = u'!prefixed'
    yaml_implicit_pattern = re.compile(re.compile(
        r'^(?P<value>[0-9]+(?:\.[0-9]+)?) (?P<prefix>[kMGTP])?(?P<unit>.*)$'))

    @classmethod
    def from_yaml(cls, loader, node):
        return PrefixedUnit(node.value)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag, unicode(data))

    def __init__(self, *args):
        if len(args) == 1:
            if isinstance(args[0], basestring):
                m = re.match(
                    r'^(?P<value>[0-9]+(?:\.[0-9]+)?) (?P<prefix>[kMGTP])?(?P<unit>.*)$', args[0])
                assert m, "Could not parse unit parameter "+repr(args[0])
                g = m.groups()
                args = [float(g[0]), g[1], g[2]]
            else:
                args = [float(args[0]), '', '']

        if len(args) == 2:
            m = re.match(r'^(?P<prefix>[kMGTP])?(?P<unit>.+)$', args[1])
            assert m, "Could not parse unit parameter"+repr(args)
            gd = m.groupdict()
            args = [float(args[0]), gd['prefix'], gd['unit']]

        if args[1] is None:
            args[1] = ''

        assert args[1] in self.PREFIXES, "Unknown prefix: "+repr(args[1])
        self.value, self.prefix, self.unit = args

    def base_value(self):
        '''gives value without prefix'''
        return self.value*self.PREFIXES[self.prefix]

    __float__ = base_value

    def good_prefix(self, max_error=0.01, round_length=2, min_prefix='', max_prefix=None):
        '''
        returns the largest prefix where the relative error is bellow *max_error* although rounded
        by *round_length*

        if *max_prefix* is found in PrefixedUnit.PREFIXES, returned value will not exceed this
        prefix.
        if *min_prefix* is given, returned value will atleast be of that prefix (no matter the
        error)
        '''
        good_prefix = min_prefix
        base_value = self.base_value()

        for k, v in self.PREFIXES.items():
            # Ignoring to large prefixes
            if max_prefix is not None and v > self.PREFIXES[max_prefix]:
                continue

            # Check that differences is < relative error *max_error*
            if abs(round(base_value/v, round_length)*v - base_value) > base_value*max_error:
                continue

            # Check that resulting number is >= 0.9
            if abs(round(base_value/v, round_length)) < 0.9:
                continue

            # Check if prefix is larger then already chosen
            if v < self.PREFIXES[good_prefix]:
                continue

            # seems to be okay
            good_prefix = k

        return good_prefix

    def with_prefix(self, prefix):
        return self.__class__(
            self.base_value()/self.PREFIXES[prefix], prefix, self.unit)

    def reduced(self):
        return self.with_prefix(self.good_prefix(max_error=0.0))

    def __str__(self):
        good_prefix = self.good_prefix()
        if self.prefix == good_prefix:
            return '{0:.2f} {1}{2}'.format(self.value, self.prefix, self.unit).strip()
        else:
            return self.with_prefix(good_prefix).__str__()

    def __repr__(self):
        return '{0}({1!r}, {2!r}, {3!r})'.format(
            self.__class__.__name__, self.value, self.prefix, self.unit)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            unit = self.unit+other.unit
        else:
            unit = self.unit

        v = self.__class__(float(self)*float(other), unit)
        return v.reduced()

    def __div__(self, other):
        if isinstance(other, self.__class__):
            unit = self.unit+'/'+other.unit
        else:
            unit = self.unit

        v = self.__class__(float(self)/float(other), unit)
        return v.reduced()

# Make this tag automatic
yaml.add_implicit_resolver(PrefixedUnit.yaml_tag, PrefixedUnit.yaml_implicit_pattern)



#!/usr/bin/env python
from __future__ import print_function

import subprocess
import re
import sys
from pprint import pprint
from itertools import chain
from copy import copy

import yaml

from prefixedunit import PrefixedUnit


def get_match_or_break(regex, haystack, flags=re.MULTILINE):
    m = re.search(regex, haystack, flags)
    assert m, "could not find "+repr(regex)
    return m.groups()


def get_machine_topology():
    topo = subprocess.Popen(['likwid-topology'], stdout=subprocess.PIPE).communicate()[0]
    machine = {
        'model name': get_match_or_break(r'^CPU type:\s+(.+?)\s*$', topo)[0],
        'sockets': int(get_match_or_break(r'^Sockets:\s+([0-9]+)\s*$', topo)[0]),
        'cores per socket': int(get_match_or_break(r'^Cores per socket:\s+([0-9]+)\s*$', topo)[0]),
        'threads per core': int(get_match_or_break(r'^Threads per core:\s+([0-9]+)\s*$', topo)[0]),
        'clock': 'INFORMATION_REQUIRED',
        'FLOPs per cycle': 'INFORMATION_REQUIRED',
        'micro-architecture': 'INFORMATION_REQUIRED',
        'icc architecture flags': 'INFORMATION_REQUIRED',
        'cacheline size': 'INFORMATION_REQUIRED',
        'overlapping ports': 'INFORAMTION_REQUIRED',
        'non-overlapping ports': 'INFORMATION_REQUIRED',
    }

    threads_start = topo.find('HWThread')
    threads_end = topo.find('Cache Topology')
    threads = {}
    for line in topo[threads_start:threads_end].split('\n'):
        m = re.match(r'([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)', line)
        if m:
            threads[m.groups()[0]] = (m.groups()[1:])

    cache_start = topo.find('Cache Topology')
    cache_end = topo.find('NUMA Topology')
    machine['memory hierarchy'] = []
    mem_level = {}
    for line in topo[cache_start:cache_end].split('\n'):
        if line.startswith('Level:'):
            mem_level['level'] = 'L'+line.split(':')[1].strip()
        elif line.startswith('Size:'):
            mem_level['size per group'] = PrefixedUnit(line.split(':')[1].strip())
        elif line.startswith('Cache groups:'):
            mem_level['groups'] = line.count('(')
            mem_level['cores per group'] = \
                (machine['cores per socket'] * machine['sockets']) / mem_level['groups']
            mem_level['threads per group'] = \
                mem_level['cores per group'] * machine['threads per core']
        mem_level['cycles per cacheline transfer'] = 'INFORMATION_REQUIRED'
        mem_level['bandwidth per core'] = 'INFORMATION_REQUIRED'
        mem_level['max. total bandwidth'] = 'INFORMATION_REQUIRED'

        if len(mem_level) == 8:
            machine['memory hierarchy'].append(mem_level)
            mem_level = {}
    machine['memory hierarchy'].append({
        'level': 'MEM',
        'size per group': None,
        'cores per group': machine['cores per socket'],
        'threads per group': machine['threads per core'] * machine['cores per socket'],
        'cycles per cacheline transfer': None,
        'bandwidth': 'INFORMATION_REQUIRED'
    })

    return machine


def measure_bw(type_, total_size, threads_per_core, max_threads_per_core, cores_per_socket,
               sockets, iterations=1000):
    """*size* is given in kilo bytes"""
    groups = []
    for s in range(sockets):
        groups += [
            '-w',
            'S' + str(s) + ':' + str(total_size) + 'kB:' +
            str(threads_per_core * cores_per_socket) +
            ':1:'+str(int(max_threads_per_core/threads_per_core))]
    # for older likwid versions add ['-g', str(sockets), '-i', str(iterations)] to cmd
    cmd = ['likwid-bench', '-t', type_]+groups
    output = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
    bw = float(get_match_or_break(r'^MByte/s:\s+([0-9]+(?:\.[0-9]+)?)\s*$', output)[0])
    return PrefixedUnit(bw, 'MB/s')


def main():
    machine = get_machine_topology()

    total_threads = machine['threads per core'] * machine['cores per socket']
    benchmarks = {'kernels': {}, 'measurements': {}}
    machine['benchmarks'] = benchmarks
    benchmarks['kernels'] = {
        'load': {
            'read streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'read+write streams': {'streams': 0, 'bytes': PrefixedUnit(0, 'B')},
            'write streams': {'streams': 0, 'bytes': PrefixedUnit(0, 'B')},
            'FLOPs per iteration': 0},
        'copy': {
            'read streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'read+write streams': {'streams': 0, 'bytes': PrefixedUnit(0, 'B')},
            'write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'FLOPs per iteration': 0},
        'update': {
            'read streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'read+write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'FLOPs per iteration': 0},
        'triad': {
            'read streams': {'streams': 3, 'bytes': PrefixedUnit(24, 'B')},
            'read+write streams': {'streams': 0, 'bytes': PrefixedUnit(0, 'B')},
            'write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'FLOPs per iteration': 2},
        'daxpy': {
            'read streams': {'streams': 2, 'bytes': PrefixedUnit(16, 'B')},
            'read+write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'write streams': {'streams': 1, 'bytes': PrefixedUnit(8, 'B')},
            'FLOPs per iteration': 2}, }

    USAGE_FACTOR = 0.5

    cores = range(1, machine['cores per socket']+1)
    for mem in machine['memory hierarchy']:
        measurement = {}
        machine['benchmarks']['measurements'][mem['level']] = measurement

        for threads_per_core in range(1, machine['threads per core']+1):
            threads = [c*threads_per_core for c in cores]
            if mem['size per group'] is not None:
                total_sizes = [
                    max(mem['size per group']*c/mem['cores per group'],
                        mem['size per group'])*USAGE_FACTOR
                    for c in cores]
            else:
                last_mem = machine['memory hierarchy'][-2]
                total_sizes = [
                    max(last_mem['size per group']*c/mem['cores per group'],
                        last_mem['size per group'])/USAGE_FACTOR
                    for c in cores]
            sizes_per_core = [t/cores[i] for i, t in enumerate(total_sizes)]
            sizes_per_thread = [t/threads[i] for i, t in enumerate(total_sizes)]

            measurement[threads_per_core] = {
                'threads per core': threads_per_core,
                'cores': copy(cores),
                'threads': threads,
                'size per core': sizes_per_core,
                'size per thread': sizes_per_thread,
                'total size': total_sizes,
                'results': {}, }

    print('Progress: ', end='', file=sys.stderr)
    sys.stderr.flush()
    for mem_level in machine['benchmarks']['measurements'].keys():
        for threads_per_core in machine['benchmarks']['measurements'][mem_level].keys():
            measurement = machine['benchmarks']['measurements'][mem_level][threads_per_core]
            measurement['results'] = {}
            for kernel in machine['benchmarks']['kernels'].keys():
                measurement['results'][kernel] = []
                for i, total_size in enumerate(measurement['total size']):
                    measurement['results'][kernel].append(measure_bw(
                        kernel,
                        int(float(total_size)/1000),
                        threads_per_core,
                        machine['threads per core'],
                        measurement['cores'][i],
                        1,  # Sockets
                        iterations=1000))

                    print('.', end='', file=sys.stderr)
                    sys.stderr.flush()

    print(yaml.dump(machine))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# pylint: disable=W0142

class Intervals:
    '''Very simple interval implementation for integers (might also work on floats)'''

    def __init__(self, *args, **kwargs):
        '''if keywords *sane* is True (default: False), checks will not be done on given data.'''
        self._data = list(args)
        if not kwargs.get('sane', False):
            self._data = [d for d in self._data if d[1] > d[0]]
            self._enforce_order()
            self._enforce_no_overlap()

    def _enforce_order(self):
        self._data.sort(key=lambda d: d[0])

    def _enforce_no_overlap(self):
        i = 0
        while i+1 < len(self._data):
            if self._data[i][1] >= self._data[i+1][0]:
                # beginning of i+1-th range is contained in i-th range
                if self._data[i][1] < self._data[i+1][1]:
                    # i+1-th range is longer, thus enlarge i-th range
                    self._data[i][1] = self._data[i+1][1]

                # removed contained range
                del self._data[i+1]
            i += 1

    def __and__(self, other):
        return Intervals(*(self._data+other._data))

    def __len__(self):
        '''returns sum of range lengths'''
        return sum(upper-lower for (lower, upper) in self._data)

    def __contains__(self, needle):
        return any(lower <= needle < upper for (lower, upper) in self._data)

    def __repr__(self):
        return str(self.__class__) + '(' + ', '.join(map(list.__repr__, self._data)) + ')'

    def __cmp__(self, other):
        if self._data == other._data:
            return 0
        else:
            return 1

if __name__ == '__main__':
    assert Intervals([0, 10])._data == [[0, 10]]
    assert Intervals([0, 10], [1, 9])._data == [[0, 10]]
    assert Intervals([0, 10], [5, 15])._data == [[0, 15]]
    assert Intervals([-5, 5], [0, 10])._data == [[-5, 10]]
    assert Intervals([0, 9], [10, 11])._data == [[0, 9], [10, 11]]
    assert Intervals([0, 10], [10, 11])._data == [[0, 11]]
    assert Intervals([0, 5]) & Intervals([1, 9]) == Intervals([0, 9])
    assert Intervals([0, 5]) & Intervals([5, 9]) == Intervals([0, 9])
    assert Intervals([2, 4]) & Intervals([0, 9]) == Intervals([0, 9])
    assert len(Intervals([1, 2])) == 1
    assert 10 not in Intervals([0, 10]) and 0 in Intervals([0, 10])



#!/usr/bin/env python

from __future__ import print_function

import re
import sys

START_MARKER = ['        movl      $111, %ebx # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
                '        .byte     100        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
                '        .byte     103        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
                '        .byte     144        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n']
END_MARKER = ['        movl      $222, %ebx # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
              '        .byte     100        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
              '        .byte     103        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n'
              '        .byte     144        # INSERTED BY KERNCRAFT IACA MARKER UTILITY\n']


def find_asm_blocks(asm_lines, with_nop=True):
    '''
    if with_nop is True, only blocks within nop markers will be considered
    '''
    blocks = []

    last_label_line = -1
    last_label = None
    packed_ctr = 0
    avx_ctr = 0
    xmm_references = []
    ymm_references = []
    gp_references = []
    last_incr = None
    within_nop_region = False
    for i, line in enumerate(asm_lines):
        if with_nop:
            if line.strip().startswith("nop"):
                within_nop_region = not within_nop_region
            if not within_nop_region:
                continue

        # Register access counts
        ymm_references += re.findall('%ymm[0-9]+', line)
        xmm_references += re.findall('%xmm[0-9]+', line)
        gp_references += re.findall('%r[a-z0-9]+', line)

        if re.match(r"^[v]?(mul|add|sub|div)[h]?p[ds]", line.strip()):
            if line.strip().startswith('v'):
                avx_ctr += 1
            packed_ctr += 1
        elif re.match(r'^\S+:', line):
            last_label = line[0:line.find(':')]
            last_label_line = i

            # Reset counters
            packed_ctr = 0
            avx_ctr = 0
            xmm_references = []
            ymm_references = []
            gp_references = []
            last_incr = None
        elif re.match(r'^inc[bwlq]?', line.strip()):
            last_incr = 1
        elif re.match(r'^add[bwlq]?\s+\$[0-9]+,', line.strip()):
            const_start = line.find('$')+1
            const_end = line[const_start+1:].find(',')+const_start+1
            last_incr = int(line[const_start:const_end])
        elif re.match(r'^dec[bwlq]?', line.strip()):
            last_incr = -1
        elif re.match(r'^sub[bwlq]?\s+\$[0-9]+,', line.strip()):
            const_start = line.find('$')+1
            const_end = line[const_start+1:].find(',')+const_start+1
            last_incr = -int(line[const_start:const_end])
        elif last_label and re.match(r'^j[a-z]+\s+'+re.escape(last_label)+r'\s+', line.strip()):
            blocks.append({'first_line': last_label_line,
                           'last_line': i,
                           'lines': i-last_label_line,
                           'label': last_label,
                           'packed_instr': packed_ctr,
                           'avx_instr': avx_ctr,
                           'XMM': (len(xmm_references), len(set(xmm_references))),
                           'YMM': (len(ymm_references), len(set(ymm_references))),
                           'GP': (len(gp_references), len(set(gp_references))),
                           'regs': (len(xmm_references) + len(ymm_references) + len(gp_references),
                                    len(set(xmm_references)) + len(set(ymm_references)) +
                                    len(set(gp_references))),
                           'loop_increment': last_incr, })

    return list(enumerate(blocks))


def select_best_block(blocks):
    # TODO make this cleverer with more stats
    best_block = max(blocks, key=lambda b: b[1]['packed_instr'])
    if best_block[1]['packed_instr'] == 0:
        best_block = max(blocks, key=lambda b: b[1]['lines'])

    return best_block[0]


def userselect_block(blocks, default=None):
    print("Blocks found in assembly file:")
    print("   block   | OPs | pck. | AVX || Registers |    YMM   |    XMM   |    GP   || l.inc |\n"
          "-----------+-----+------+-----++-----------+----------+----------+---------++-------|")
    for idx, b in blocks:
        print('{:>2} {b[label]:>5} | {b[lines]:>3} | {b[packed_instr]:>4} | {b[avx_instr]:>3} |'
              '| {b[regs][0]:>3} ({b[regs][1]:>3}) | {b[YMM][0]:>3} ({b[YMM][1]:>2}) | '
              '{b[XMM][0]:>3} ({b[XMM][1]:>2}) | {b[GP][0]:>2} ({b[GP][1]:>2}) || '
              '{b[loop_increment]:>5} |'.format(idx, b=b))

    # Let user select block:
    block_idx = -1
    while not (0 <= block_idx < len(blocks)):
        block_idx = raw_input("Choose block to be marked ["+str(default)+"]: ") or default
        try:
            block_idx = int(block_idx)
        except ValueError:
            block_idx = -1
    block = blocks[block_idx][1]

    return block_idx


def insert_markers(asm_lines, start_line, end_line):
    asm_lines = asm_lines[:start_line] + START_MARKER + \
        asm_lines[start_line:end_line+1] + END_MARKER + \
        asm_lines[end_line+1:]
    return asm_lines


def main():
    if len(sys.argv) != 2:
        print("Usage:", sys.argv[0], "filename.s")
        sys.exit(1)

    with open(sys.argv[1], 'r') as fp:
        lines = fp.readlines()
    blocks = find_asm_blocks(lines)

    # TODO check for already present markers

    # Choose best default block:
    best_idx = select_best_block(blocks)
    # Let user select block:
    block_idx = userselect_block(blocks, best_idx)

    block = blocks[block_idx][1]

    # Insert markers:
    lines = insert_markers(lines, block['first_line'], block['last_line'])

    # write back to file
    with open(sys.argv[1], 'w') as fp:
        fp.writelines(lines)

    print("Markers inserted.")

if __name__ == '__main__':
    main()



'''Collection of performance models

This model combines all performance models currently supported by kerncraft. Only the performace
model class is exported, so please add new models to __all__.

The exported classes must have the following class level attributes:
  * name (str) is the name of the performance model (no abreviatation)
  * configure_subparser(parser) classmethod that configures the parser for cli usage
  * construct_from_args(kernel, machine, args) classmethod that construct the object
  * analyze() that analyses ther kernel with regard to the machine definition and args passed
  * report() return a readable text output with analysis report
  * results (dict) must be available after analyze has been called
'''
from ecm import ECM, ECMData, ECMCPU
from roofline import Roofline, RooflineIACA

__all__ = ['ECM', 'ECMData', 'ECMCPU', 'Roofline', 'RooflineIACA']



#!/usr/bin/env python
# pylint: disable=W0142

from __future__ import print_function

from copy import copy
import operator
import tempfile
import subprocess
import os.path

from pycparser import CParser, c_ast, c_generator
from pycparser.c_generator import CGenerator

import iaca_marker as iaca

def prefix_indent(prefix, textblock, later_prefix=' '):
    textblock = textblock.split('\n')
    s = prefix + textblock[0] + '\n'
    if len(later_prefix) == 1:
        later_prefix = ' '*len(prefix)
    s = s+'\n'.join(map(lambda x: later_prefix+x, textblock[1:]))
    if s[-1] != '\n':
        return s + '\n'
    else:
        return s


def trasform_multidim_to_1d_decl(decl):
    '''
    Transforms ast of multidimensional declaration to a single dimension declaration.
    In-place operation!

    Returns name and dimensions of array (to be used with transform_multidim_to_1d_ref())
    '''
    dims = []
    t = decl.type
    while type(t) is c_ast.ArrayDecl:
        dims.append(t.dim)
        t = t.type

    if dims:
        # Multidimensional array
        decl.type.dim = reduce(lambda l, r: c_ast.BinaryOp('*', l, r), dims)
        decl.type.type = t

    return decl.name, dims


def transform_multidim_to_1d_ref(aref, dimension_dict):
    '''
    Transforms ast of multidimensional reference to a single dimension reference.
    In-place operation!
    '''
    dims = []
    name = aref
    while type(name) is c_ast.ArrayRef:
        dims.append(name.subscript)
        name = name.name

    subscript_list = []
    for i, d in enumerate(dims):

        if i == 0:
            subscript_list.append(d)
        else:
            subscript_list.append(c_ast.BinaryOp('*', d, reduce(
                lambda l, r: c_ast.BinaryOp('*', l, r),
                dimension_dict[name.name][-i-1:-1])))

    aref.subscript = reduce(
        lambda l, r: c_ast.BinaryOp('+', l, r), subscript_list)
    aref.name = name


def transform_array_decl_to_malloc(decl):
    '''Trans forms "type var_name[N]" to "type* var_name = malloc(sizeof(type)*N)'''
    if type(decl.type) is not c_ast.ArrayDecl:
        # Not an array declaration, can be ignored
        return

    type_ = c_ast.PtrDecl([], decl.type.type)
    init = c_ast.FuncCall(
        c_ast.ID('malloc'),
        c_ast.ExprList([
            c_ast.BinaryOp(
                '*',
                c_ast.UnaryOp(
                    'sizeof',
                    c_ast.Typename([], c_ast.TypeDecl(None, [], decl.type.type.type))),
                decl.type.dim)]))

    decl.type = type_
    decl.init = init


def find_array_references(ast):
    '''returns list of array references in AST'''
    if type(ast) is c_ast.ArrayRef:
        return [ast]
    else:
        return reduce(operator.add, map(lambda o: find_array_references(o[1]), ast.children()), [])


class Kernel:
    def __init__(self, kernel_code, constants=None, variables=None, filename=None):
        '''This class captures the DSL kernel code, analyzes it and reports access pattern'''
        self.kernel_code = kernel_code
        self._filename = filename

        parser = CParser()
        self.kernel_ast = parser.parse(self.as_function()).ext[0].body

        self._loop_stack = []
        self._sources = {}
        self._destinations = {}
        self._constants = {} if constants is None else constants
        self._variables = {} if variables is None else variables
        self._flops = {}
        self.blocks = {}  # ASM block information, populated after call to assemble
        self.block_idx = None  # Block index used for marking (and containing the inner loop)

    def as_function(self, func_name='test'):
        return 'void {}() {{ {} }}'.format(func_name, self.kernel_code)

    def set_constant(self, name, value):
        assert type(name) is str, "constant name needs to be of type str"
        assert type(value) is int, "constant value needs to be of type int"
        self._constants[name] = value

    def set_variable(self, name, type_, size):
        assert type_ in ['double', 'float'], 'only float and double variables are supported'
        assert type(size) in [tuple, type(None)], 'size has to be defined as tuple'
        self._variables[name] = (type_, size)

    def process(self):
        assert type(self.kernel_ast) is c_ast.Compound, "Kernel has to be a compound statement"
        assert all(map(lambda s: type(s) is c_ast.Decl, self.kernel_ast.block_items[:-1])), \
            'all statments befor the for loop need to be declarations'
        assert type(self.kernel_ast.block_items[-1]) is c_ast.For, \
            'last statment in kernel code must be a loop'

        for item in self.kernel_ast.block_items[:-1]:
            array = type(item.type) is c_ast.ArrayDecl

            if array:
                dims = []
                t = item.type
                while type(t) is c_ast.ArrayDecl:
                    dims.append(self.conv_ast_to_int(t.dim))
                    t = t.type

                assert len(t.type.names) == 1, "only single types are supported"
                self.set_variable(item.name, t.type.names[0], tuple(dims))

            else:
                assert len(item.type.type.names) == 1, "only single types are supported"
                self.set_variable(item.name, item.type.type.names[0], None)

        floop = self.kernel_ast.block_items[-1]
        self._p_for(floop)

    def conv_ast_to_int(self, math_ast):
        '''
        converts mathematical expressions containing paranthesis, addition, subtraction and
        multiplication from AST to int.
        '''
        if type(math_ast) is c_ast.ID:
            return self._constants[math_ast.name]
        elif type(math_ast) is c_ast.Constant:
            return int(math_ast.value)
        else:  # elif type(dim) is c_ast.BinaryOp:
            op = {
                '*': operator.mul,
                '+': operator.add,
                '-': operator.sub
            }

            return int(op[math_ast.op](
                self.conv_ast_to_int(math_ast.left),
                self.conv_ast_to_int(math_ast.right)))

    def _get_offsets(self, aref, dim=0):
        '''
        returns a list of offsets of an ArrayRef object in all dimensions

        the index order is right to left (c-code order).
        e.g. c[i+1][j-2] -> [-2, +1]
        '''

        # Check for restrictions
        assert type(aref.name) in [c_ast.ArrayRef, c_ast.ID], \
            "array references must only be used with variables or other array references"
        assert type(aref.subscript) in [c_ast.ID, c_ast.Constant, c_ast.BinaryOp], \
            'array subscript must only contain variables or binary operations'

        idxs = []

        # TODO work-in-progress generisches auswerten von allem in [...]
        if type(aref.subscript) is c_ast.BinaryOp:
            assert aref.subscript.op in '+-', \
                'binary operations in array subscript must by + or -'
            assert (type(aref.subscript.left) is c_ast.ID and
                    type(aref.subscript.right) is c_ast.Constant), \
                'binary operation in array subscript may only have form "variable +- constant"'
            assert aref.subscript.left.name in map(lambda l: l[0], self._loop_stack), \
                'varialbes used in array indices has to be a loop counter'

            sign = 1 if aref.subscript.op == '+' else -1
            offset = sign*int(aref.subscript.right.value)

            idxs.append(('rel', aref.subscript.left.name, offset))
        elif type(aref.subscript) is c_ast.ID:
            assert aref.subscript.name in map(lambda l: l[0], self._loop_stack), \
                'varialbes used in array indices has to be a loop counter'
            idxs.append(('rel', aref.subscript.name, 0))
        else:  # type(aref.subscript) is c_ast.Constant
            idxs.append(('abs', int(aref.subscript.value)))

        if type(aref.name) is c_ast.ArrayRef:
            idxs += self._get_offsets(aref.name, dim=dim+1)

        if dim == 0:
            idxs.reverse()

        return idxs

    @classmethod
    def _get_basename(cls, aref):
        '''
        returns base name of ArrayRef object

        e.g. c[i+1][j-2] -> 'c'
        '''

        if type(aref.name) is c_ast.ArrayRef:
            return cls._get_basename(aref.name)
        else:
            return aref.name.name

    def _p_for(self, floop):
        # Check for restrictions
        assert type(floop) is c_ast.For, "May only be a for loop"
        assert hasattr(floop, 'init') and hasattr(floop, 'cond') and hasattr(floop, 'next'), \
            "Loop must have initial, condition and next statements."
        assert type(floop.init) is c_ast.DeclList, \
            "Initialization of loops need to be declarations."
        assert len(floop.init.decls) == 1, "Only single declaration is allowed in init. of loop."
        assert floop.cond.op in '<', "only lt (<) is allowed as loop condition"
        assert type(floop.cond.left) is c_ast.ID, 'left of cond. operand has to be a variable'
        assert type(floop.cond.right) in [c_ast.Constant, c_ast.ID, c_ast.BinaryOp], \
            'right of cond. operand has to be a constant, a variable or a binary operation'
        assert type(floop.next) in [c_ast.UnaryOp, c_ast.Assignment], \
            'next statement has to be a unary or assignment operation'
        assert floop.next.op in ['++', 'p++', '+='], 'only ++ and += next operations are allowed'
        assert type(floop.stmt) in [c_ast.Compound, c_ast.Assignment, c_ast.For], \
            'the inner loop may contain only assignments or compounds of assignments'

        if type(floop.cond.right) is c_ast.ID:
            const_name = floop.cond.right.name
            assert const_name in self._constants, \
                'loop right operand has to be defined as a constant in ECM object'
            iter_max = self._constants[const_name]
        elif type(floop.cond.right) is c_ast.Constant:
            iter_max = int(floop.cond.right.value)
        else:  # type(floop.cond.right) is c_ast.BinaryOp
            bop = floop.cond.right
            assert type(bop.left) is c_ast.ID, 'left of operator has to be a variable'
            assert type(bop.right) is c_ast.Constant, 'right of operator has to be a constant'
            assert bop.op in '+-', 'only plus (+) and minus (-) are accepted operators'

            sign = 1 if bop.op == '+' else -1
            iter_max = self._constants[bop.left.name]+sign*int(bop.right.value)

        if type(floop.next) is c_ast.Assignment:
            assert type(floop.next.lvalue) is c_ast.ID, \
                'next operation may only act on loop counter'
            assert type(floop.next.rvalue) is c_ast.Constant, 'only constant increments are allowed'
            assert floop.next.lvalue.name == floop.cond.left.name == floop.init.decls[0].name, \
                'initial, condition and next statement of for loop must act on same loop ' \
                'counter variable'
            step_size = int(floop.next.rvalue.value)
        else:
            assert type(floop.next.expr) is c_ast.ID, 'next operation may only act on loop counter'
            assert floop.next.expr.name == floop.cond.left.name == floop.init.decls[0].name, \
                'initial, condition and next statement of for loop must act on same loop ' \
                'counter variable'
            step_size = 1

        # Document for loop stack
        self._loop_stack.append(
            # (index name, min, max, step size)
            (floop.init.decls[0].name, floop.init.decls[0].init.value, iter_max, step_size)
        )
        # TODO add support for other stepsizes (even negative/reverse steps?)

        # Traverse tree
        if type(floop.stmt) is c_ast.For:
            self._p_for(floop.stmt)
        elif type(floop.stmt) is c_ast.Compound and \
                len(floop.stmt.block_items) == 1 and \
                type(floop.stmt.block_items[0]) is c_ast.For:
            self._p_for(floop.stmt.block_items[0])
        elif type(floop.stmt) is c_ast.Assignment:
            self._p_assignment(floop.stmt)
        else:  # type(floop.stmt) is c_ast.Compound
            for assgn in floop.stmt.block_items:
                self._p_assignment(assgn)

    def _p_assignment(self, stmt):
        # Check for restrictions
        assert type(stmt) is c_ast.Assignment, \
            "Only assignment statements are allowed in loops."
        assert type(stmt.lvalue) in [c_ast.ArrayRef, c_ast.ID], \
            "Only assignment to array element or varialbe is allowed."

        write_and_read = False
        if stmt.op != '=':
            write_and_read = True
            op = stmt.op.strip('=')
            self._flops[op] = self._flops.get(op, 0)+1

        # Document data destination
        if type(stmt.lvalue) is c_ast.ArrayRef:
            # self._destinations[dest name] = [dest offset, ...])
            self._destinations.setdefault(self._get_basename(stmt.lvalue), [])
            self._destinations[self._get_basename(stmt.lvalue)].append(
                self._get_offsets(stmt.lvalue))

            if write_and_read:
                # this means that +=, -= or something of that sort was used
                self._sources.setdefault(self._get_basename(stmt.lvalue), [])
                self._sources[self._get_basename(stmt.lvalue)].append(
                    self._get_offsets(stmt.lvalue))

        else:  # type(stmt.lvalue) is c_ast.ID
            self._destinations.setdefault(stmt.lvalue.name, [])
            self._destinations[stmt.lvalue.name].append([('dir',)])

            if write_and_read:
                # this means that +=, -= or something of that sort was used
                self._sources.setdefault(stmt.lvalue.name, [])
                self._sources[stmt.lvalue.name].append([('dir',)])

        # Traverse tree
        self._p_sources(stmt.rvalue)

    def _p_sources(self, stmt):
        sources = []

        assert type(stmt) in [c_ast.ArrayRef, c_ast.Constant, c_ast.ID, c_ast.BinaryOp], \
            'only references to arrays, constants and variables as well as binary operations ' + \
            'are supported'

        if type(stmt) is c_ast.ArrayRef:
            # Document data source
            bname = self._get_basename(stmt)
            self._sources.setdefault(bname, [])
            self._sources[bname].append(self._get_offsets(stmt))
            # TODO deactivated for now, since that notation might be useless
            # ArrayAccess(stmt, array_info=self._variables[bname])
        elif type(stmt) is c_ast.ID:
            # Document data source
            self._sources.setdefault(stmt.name, [])
            self._sources[stmt.name].append([('dir',)])
        elif type(stmt) is c_ast.BinaryOp:
            # Traverse tree
            self._p_sources(stmt.left)
            self._p_sources(stmt.right)

            self._flops[stmt.op] = self._flops.get(stmt.op, 0)+1

        return sources

    def as_code(self):
        ast = copy(self.kernel_ast)
        declarations = filter(lambda d: type(d) is c_ast.Decl, ast.block_items)

        # transform multi-dimensional declarations to one dimensional references
        array_dimensions = dict(map(trasform_multidim_to_1d_decl, declarations))
        # transform to pointer and malloc notation (stack can be too small)
        map(transform_array_decl_to_malloc, declarations)

        # add declarations for constants
        for k, v in self._constants.items():
            # cont int N = atoi(argv[1])
            # TODO change subscript of argv depending on constant count
            type_decl = c_ast.TypeDecl(k, ['const'], c_ast.IdentifierType(['int']))
            init = c_ast.FuncCall(
                c_ast.ID('atoi'),
                c_ast.ExprList([c_ast.ArrayRef(c_ast.ID('argv'), c_ast.Constant('int', '1'))]))
            ast.block_items.insert(0, c_ast.Decl(
                k, ['const'], [], [],
                type_decl, init, None))

        # inject array initialization
        for d in declarations:
            i = ast.block_items.index(d)

            # Build ast to inject
            if array_dimensions[d.name]:
                # this is an array, we need a for loop to initialize it
                # for(init; cond; next) stmt

                # Init: int i = 0;
                counter_name = 'i'
                while counter_name in array_dimensions:
                    counter_name = chr(ord(counter_name)+1)

                init = c_ast.DeclList([
                    c_ast.Decl(
                        counter_name, [], [], [], c_ast.TypeDecl(
                            counter_name, [], c_ast.IdentifierType(['int'])),
                        c_ast.Constant('int', '0'),
                        None)],
                    None)

                # Cond: i < ... (... is length of array)
                cond = c_ast.BinaryOp(
                    '<',
                    c_ast.ID(counter_name),
                    reduce(lambda l, r: c_ast.BinaryOp('*', l, r), array_dimensions[d.name]))

                # Next: i++
                next_ = c_ast.UnaryOp('++', c_ast.ID(counter_name))

                # Statement
                stmt = c_ast.Assignment(
                    '=',
                    c_ast.ArrayRef(c_ast.ID(d.name), c_ast.ID(counter_name)),
                    c_ast.Constant('int', '0'))

                ast.block_items.insert(i+1, c_ast.For(init, cond, next_, stmt))

                # inject dummy access to arrays, so compiler does not over-optimize code
                # TODO put if around it, so code will actually run
                ast.block_items.insert(
                    i+2, c_ast.FuncCall(c_ast.ID('dummy'), c_ast.ExprList([c_ast.ID(d.name)])))
            else:
                # this is a scalar, so a simple Assignment is enough
                ast.block_items.insert(
                    i+1, c_ast.Assignment('=', c_ast.ID(d.name), c_ast.Constant('int', '0')))

                # inject dummy access to scalar, so compiler does not over-optimize code
                # TODO put if around it, so code will actually run
                ast.block_items.insert(
                    i+2,
                    c_ast.FuncCall(c_ast.ID('dummy'),
                                   c_ast.ExprList([c_ast.UnaryOp('&', c_ast.ID(d.name))])))

        # transform multi-dimensional array references to one dimensional references
        map(lambda aref: transform_multidim_to_1d_ref(aref, array_dimensions),
            find_array_references(ast))

        # Mark the outer for-loop by injecting asm("nop");
        ast.block_items.insert(
            -1,
            c_ast.FuncCall(c_ast.ID('asm'), c_ast.ExprList([c_ast.Constant('string', '"nop"')])))
        ast.block_items.append(
            c_ast.FuncCall(c_ast.ID('asm'), c_ast.ExprList([c_ast.Constant('string', '"nop"')])))

        # embedd Compound into main FuncDecl
        decl = c_ast.Decl('main', [], [], [], c_ast.FuncDecl(c_ast.ParamList([
            c_ast.Typename([], c_ast.TypeDecl('argc', [], c_ast.IdentifierType(['int']))),
            c_ast.Typename([], c_ast.PtrDecl([], c_ast.PtrDecl(
                [], c_ast.TypeDecl('argv', [], c_ast.IdentifierType(['char'])))))]),
            c_ast.TypeDecl('main', [], c_ast.IdentifierType(['int']))),
            None, None)

        ast = c_ast.FuncDef(decl, None, ast)

        # embedd Compound AST into FileAST
        ast = c_ast.FileAST([ast])

        # add dummy function declaration
        decl = c_ast.Decl('dummy', [], [], [], c_ast.FuncDecl(
            c_ast.ParamList([c_ast.Typename([], c_ast.PtrDecl(
                [], c_ast.TypeDecl(None, [], c_ast.IdentifierType(['double']))))]),
            c_ast.TypeDecl('dummy', [], c_ast.IdentifierType(['void']))),
            None, None)

        ast.ext.insert(0, decl)

        # convert to code string
        code = CGenerator().visit(ast)

        # add "#include"s for dummy and stdlib (for malloc)
        code = '#include <stdlib.h>\n\n' + code

        return code

    def assemble(self, in_filename, out_filename=None, iaca_markers=True, asm_block='auto'):
        '''
        Assembles *in_filename* to *out_filename*.

        If *out_filename* is not given a new file will created either temporarily or according
        to kernel file location.

        if *iaca_marked* is set to true, markers are inserted around the block with most packed
        instructions or (if no packed instr. were found) the largest block and modified file is
        saved to *in_file*.

        *asm_block* controlls how the to-be-marked block is chosen. "auto" (default) results in
        the largest block, "manual" results in interactive and a number in the according block.

        Returns two-tuple (filepointer, filename) to temp binary file.
        '''
        if not out_filename:
            suffix = ''
            if iaca_markers:
                suffix += '.iaca_marked'
            if self._filename:
                out_filename = os.path.abspath(os.path.splitext(self._filename)[0]+suffix)
            else:
                out_filename = tempfile.mkstemp(suffix=suffix)

        # insert iaca markers
        if iaca_markers:
            with open(in_filename, 'r') as in_file:
                lines = in_file.readlines()
            self.blocks = iaca.find_asm_blocks(lines)

            # TODO check for already present markers

            # Choose best default block:
            block_idx = iaca.select_best_block(self.blocks)
            if asm_block == 'manual':
                block_idx = iaca.userselect_block(self.blocks, default=block_idx)
            elif asm_block != 'auto':
                block_idx = asm_block
            self.block_idx = block_idx

            block = self.blocks[block_idx][1]

            # Insert markers:
            lines = iaca.insert_markers(lines, block['first_line'], block['last_line'])

            # write back to file
            with open(in_filename, 'w') as in_file:
                in_file.writelines(lines)

        try:
            # Assamble all to a binary
            subprocess.check_output(
                ["icc", os.path.basename(in_file.name), 'dummy.s', '-o', out_filename],
                cwd=os.path.dirname(in_file.name))
        finally:
            in_file.close()

        return out_filename

    def compile(self, compiler_args=None):
        '''
        Compiles source (from as_code()) to assembly.

        Returns two-tuple (filepointer, filename) to assembly file.

        Output can be used with Kernel.assemble()
        '''

        if not self._filename:
            in_file = tempfile.NamedTemporaryFile(suffix='_compilable.c').file
        else:
            in_file = open(self._filename+"_compilable.c", 'w')

        in_file.write(self.as_code())
        in_file.flush()

        if compiler_args is None:
            compiler_args = []
        compiler_args += ['-O3', '-fno-alias', '-std=c99']

        try:
            subprocess.check_output(
                ["icc"]+compiler_args+[os.path.basename(in_file.name), '-S'],
                cwd=os.path.dirname(in_file.name))

            subprocess.check_output(
                ["icc"] + compiler_args + [
                    os.path.abspath(os.path.dirname(__file__)+'/headers/dummy.c'),
                    '-S'],
                cwd=os.path.dirname(in_file.name))
        finally:
            in_file.close()

        # Let's return the out_file name
        return os.path.splitext(in_file.name)[0]+'.s'

    def print_kernel_info(self):
        table = ('     idx |        min        max       step\n' +
                 '---------+---------------------------------\n')
        for l in self._loop_stack:
            table += '{:>8} | {:>10} {:>10} {:>+10}\n'.format(*l)
        print(prefix_indent('loop stack:        ', table))

        table = ('    name |  offsets   ...\n' +
                 '---------+------------...\n')
        for name, offsets in self._sources.items():
            prefix = '{:>8} | '.format(name)
            right_side = '\n'.join(map(lambda o: ', '.join(map(tuple.__repr__, o)), offsets))
            table += prefix_indent(prefix, right_side, later_prefix='         | ')
        print(prefix_indent('data sources:      ', table))

        table = ('    name |  offsets   ...\n' +
                 '---------+------------...\n')
        for name, offsets in self._destinations.items():
            prefix = '{:>8} | '.format(name)
            right_side = '\n'.join(map(lambda o: ', '.join(map(tuple.__repr__, o)), offsets))
            table += prefix_indent(prefix, right_side, later_prefix='         | ')
        print(prefix_indent('data destinations: ', table))

        table = (' op | count \n' +
                 '----+-------\n')
        for op, count in self._flops.items():
            table += '{:>3} | {:>4}\n'.format(op, count)
        table += '     =======\n'
        table += '      {:>4}'.format(sum(self._flops.values()))
        print(prefix_indent('FLOPs:     ', table))

    def print_kernel_code(self):
        print(self.kernel_code)

    def print_variables_info(self):
        table = ('    name |   type size             \n' +
                 '---------+-------------------------\n')
        for name, var_info in self._variables.items():
            table += '{:>8} | {:>6} {:<10}\n'.format(name, var_info[0], var_info[1])
        print(prefix_indent('variables: ', table))

    def print_constants_info(self):
        table = ('    name | value     \n' +
                 '---------+-----------\n')
        for name, value in self._constants.items():
            table += '{:>8} | {:<10}\n'.format(name, value)
        print(prefix_indent('constants: ', table))



#!/usr/bin/env python

from __future__ import print_function
from pprint import pprint
from functools import reduce
import operator
import subprocess
import re

from kerncraft.intervals import Intervals
from kerncraft.prefixedunit import PrefixedUnit

# Datatype sizes in bytes
datatype_size = {'double': 8, 'float': 4}


def blocking(indices, block_size, initial_boundary=0):
    '''
    splits list of integers into blocks of block_size. returns block indices.

    first block element is located at initial_boundary (default 0).

    >>> blocking([0, -1, -2, -3, -4, -5, -6, -7, -8, -9], 8)
    [0,-1]
    >>> blocking([0], 8)
    [0]
    '''
    blocks = []

    for idx in indices:
        bl_idx = (idx-initial_boundary)/block_size
        if bl_idx not in blocks:
            blocks.append(bl_idx)
    blocks.sort()

    return blocks


def flatten_dict(d):
    '''
    transforms 2d-dict d[i][k] into a new 1d-dict e[(i,k)] with 2-tuple keys
    '''
    e = {}
    for k in d.keys():
        for l in d[k].keys():
            e[(k, l)] = d[k][l]
    return e


class Roofline:
    """
    class representation of the Roofline Model

    more info to follow...
    """

    name = "Roofline"

    @classmethod
    def configure_arggroup(cls, parser):
        pass

    def __init__(self, kernel, machine, args=None, parser=None):
        """
        *kernel* is a Kernel object
        *machine* describes the machine (cpu, cache and memory) characteristics
        *args* (optional) are the parsed arguments from the comand line
        """
        self.kernel = kernel
        self.machine = machine
        self._args = args
        self._parser = parser

        if args:
            # handle CLI info
            pass

    def _calculate_relative_offset(self, name, access_dimensions):
        '''
        returns the offset from the iteration center in number of elements and the order of indices
        used in access.
        '''
        offset = 0
        base_dims = self.kernel._variables[name][1]

        for dim, offset_info in enumerate(access_dimensions):
            offset_type, idx_name, dim_offset = offset_info
            assert offset_type == 'rel', 'Only relative access to arrays is supported at the moment'

            if offset_type == 'rel':
                offset += dim_offset*reduce(operator.mul, base_dims[dim+1:], 1)
            else:
                # should not happen
                pass

        return offset

    def _calculate_iteration_offset(self, name, index_order, loop_index):
        '''
        returns the offset from one to the next iteration using *loop_index*.
        *index_order* is the order used by the access dimensions e.g. 'ijk' corresponse to [i][j][k]
        *loop_index* specifies the loop to be used for iterations (this is typically the inner
        moste one)
        '''
        offset = 0
        base_dims = self.kernel._variables[name][1]

        for dim, index_name in enumerate(index_order):
            if loop_index == index_name:
                offset += reduce(operator.mul, base_dims[dim+1:], 1)

        return offset

    def _get_index_order(self, access_dimensions):
        '''Returns the order of indices used in *access_dimensions*.'''
        return ''.join(map(lambda d: d[1], access_dimensions))

    def _expand_to_cacheline_blocks(self, first, last):
        '''
        Returns first and last values wich align with cacheline blocks, by increasing range.
        '''
        # TODO how to handle multiple datatypes (with different size)?
        element_size = datatype_size['double']
        elements_per_cacheline = int(float(self.machine['cacheline size'])) / element_size

        first = first - first % elements_per_cacheline
        last = last - last % elements_per_cacheline + elements_per_cacheline - 1

        return [first, last]

    def calculate_cache_access(self):
        results = {'bottleneck level': None, 'mem bottlenecks': []}

        read_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}
        write_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}
        iteration_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}

        # TODO how to handle multiple datatypes (with different size)?
        element_size = datatype_size['double']
        elements_per_cacheline = int(float(self.machine['cacheline size'])) / element_size

        loop_order = ''.join(map(lambda l: l[0], self.kernel._loop_stack))

        for var_name in self.kernel._variables.keys():
            var_type, var_dims = self.kernel._variables[var_name]

            # Skip the following access: (they are hopefully kept in registers)
            #   - scalar values
            if var_dims is None:
                continue
            #   - access does not change with inner-most loop index
            writes = filter(
                lambda acs: loop_order[-1] in map(lambda a: a[1], acs),
                self.kernel._destinations.get(var_name, []))
            reads = filter(
                lambda acs: loop_order[-1] in map(lambda a: a[1], acs),
                self.kernel._sources.get(var_name, []))

            # Compile access pattern
            for r in reads:
                offset = self._calculate_relative_offset(var_name, r)
                idx_order = self._get_index_order(r)
                read_offsets[var_name].setdefault(idx_order, [])
                read_offsets[var_name][idx_order].append(offset)
            for w in writes:
                offset = self._calculate_relative_offset(var_name, w)
                idx_order = self._get_index_order(w)
                write_offsets[var_name].setdefault(idx_order, [])
                write_offsets[var_name][idx_order].append(offset)

            # With ECM we would do unrolling, but not with roofline

        # initialize misses and hits
        misses = {}
        hits = {}
        evicts = {}
        total_misses = {}
        total_hits = {}
        total_evicts = {}
        total_lines_misses = {}
        total_lines_hits = {}
        total_lines_evicts = {}

        # Check for layer condition towards all cache levels (except main memory/last level)
        for cache_level, cache_info in list(enumerate(self.machine['memory hierarchy']))[:-1]:
            cache_size = int(float(cache_info['size per group']))
            cache_cycles = cache_info['cycles per cacheline transfer']

            trace_length = 0
            updated_length = True
            while updated_length:
                updated_length = False

                # Initialize cache, misses, hits and evicts for current level
                cache = {}
                misses[cache_level] = {}
                hits[cache_level] = {}
                evicts[cache_level] = {}

                # We consider everythin a miss in the beginning
                # TODO here read and writes are treated the same, this implies write-allocate
                #      to support nontemporal stores, this needs to be changed
                for name in read_offsets.keys()+write_offsets.keys():
                    cache[name] = {}
                    misses[cache_level][name] = {}
                    hits[cache_level][name] = {}

                    for idx_order in read_offsets[name].keys()+write_offsets[name].keys():
                        cache[name][idx_order] = Intervals()
                        if cache_level-1 not in misses:
                            misses[cache_level][name][idx_order] = sorted(
                                read_offsets.get(name, {}).get(idx_order, []) +
                                write_offsets.get(name, {}).get(idx_order, []),
                                reverse=True)
                        else:
                            misses[cache_level][name][idx_order] = list(
                                misses[cache_level-1][name][idx_order])
                        hits[cache_level][name][idx_order] = []

                # Caches are still empty (thus only misses)
                trace_count = 0
                cache_used_size = 0

                # Now we trace the cache access backwards (in time/iterations) and check for hits
                for var_name in misses[cache_level].keys():
                    for idx_order in misses[cache_level][var_name].keys():
                        iter_offset = self._calculate_iteration_offset(
                            var_name, idx_order, loop_order[-1])

                        # Add cache trace
                        for offset in list(misses[cache_level][var_name][idx_order]):
                            # If already present in cache add to hits
                            if offset in cache[var_name][idx_order]:
                                misses[cache_level][var_name][idx_order].remove(offset)

                                # We might have multiple hits on the same offset (e.g in DAXPY)
                                if offset not in hits[cache_level][var_name][idx_order]:
                                    hits[cache_level][var_name][idx_order].append(offset)

                            # Add cache, we can do this since misses are sorted in reverse order of
                            # access and we assume LRU cache replacement policy
                            if iter_offset <= elements_per_cacheline:
                                # iterations overlap, thus we can savely add the whole range
                                cached_first, cached_last = self._expand_to_cacheline_blocks(
                                    offset-iter_offset*trace_length, offset+1)
                                cache[var_name][idx_order] &= Intervals(
                                    [cached_first, cached_last+1], sane=True)
                            else:
                                # There is no overlap, we can append the ranges onto one another
                                # TODO optimize this code section (and maybe merge with above)
                                new_cache = [self._expand_to_cacheline_blocks(o, o) for o in range(
                                    offset-iter_offset*trace_length, offset+1, iter_offset)]
                                new_cache = Intervals(*new_cache, sane=True)
                                cache[var_name][idx_order] &= new_cache

                        trace_count += len(cache[var_name][idx_order]._data)
                        cache_used_size += len(cache[var_name][idx_order])*element_size

                # Calculate new possible trace_length according to free space in cache
                # TODO take CL blocked access into account
                # TODO make /2 customizable
                new_trace_length = trace_length + \
                    ((cache_size/2 - cache_used_size)/trace_count)/element_size

                if new_trace_length > trace_length:
                    trace_length = new_trace_length
                    updated_length = True

                # All writes to require the data to be evicted eventually
                evicts[cache_level] = \
                    {var_name: dict() for var_name in self.kernel._variables.keys()}
                for name in write_offsets.keys():
                    for idx_order in write_offsets[name].keys():
                        evicts[cache_level][name][idx_order] = list(write_offsets[name][idx_order])

            # Compiling stats
            total_misses[cache_level] = sum(map(
                lambda l: sum(map(len, l.values())),
                misses[cache_level].values()))
            total_hits[cache_level] = sum(map(
                lambda l: sum(map(len, l.values())),
                hits[cache_level].values()))
            total_evicts[cache_level] = sum(map(
                lambda l: sum(map(len, l.values())),
                evicts[cache_level].values()))

            total_lines_misses[cache_level] = sum(map(
                lambda o: sum(map(lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                misses[cache_level].values()))
            total_lines_hits[cache_level] = sum(map(
                lambda o: sum(map(lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                hits[cache_level].values()))
            total_lines_evicts[cache_level] = sum(map(
                lambda o: sum(map(lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                evicts[cache_level].values()))

            # Calculate performance (arithmetic intensity * bandwidth with
            # arithmetic intensity = flops / bytes transfered)
            bytes_transfered = (total_misses[cache_level]+total_evicts[cache_level])*element_size
            total_flops = sum(self.kernel._flops.values())
            arith_intens = float(total_flops)/float(bytes_transfered)

            # choose bw according to cache level and problem
            # first, compile stream counts at current cache level
            # write-allocate is allready resolved above
            read_streams = 0
            for var_name in misses[cache_level].keys():
                for idx_order in misses[cache_level][var_name]:
                    read_streams += len(misses[cache_level][var_name][idx_order])
            write_streams = 0
            for var_name in evicts[cache_level].keys():
                for idx_order in evicts[cache_level][var_name]:
                    write_streams += len(evicts[cache_level][var_name][idx_order])
            # second, try to find best fitting kernel (closest to stream seen stream counts):
            # write allocate has to be handled in kernel information (all writes are also reads)
            # TODO support for non-write-allocate architectures
            measurement_kernel = 'load'
            measurement_kernel_info = self.machine['benchmarks']['kernels'][measurement_kernel]
            for kernel_name, kernel_info in self.machine['benchmarks']['kernels'].items():
                if (read_streams >= (kernel_info['read streams']['streams'] +
                                     kernel_info['write streams']['streams'] -
                                     kernel_info['read+write streams']['streams']) >
                        measurement_kernel_info['read streams']['streams'] +
                        measurement_kernel_info['write streams']['streams'] -
                        measurement_kernel_info['read+write streams']['streams'] and
                        write_streams >= kernel_info['write streams']['streams'] >
                        measurement_kernel_info['write streams']['streams']):
                    measurement_kernel = kernel_name
                    measurement_kernel_info = kernel_info

            # TODO choose smt and cores:
            threads_per_core, cores = 1, 1
            bw_measurements = \
                self.machine['benchmarks']['measurements'][cache_info['level']][threads_per_core]
            assert threads_per_core == bw_measurements['threads per core'], \
                'malformed measurement dictionary in machine file.'
            run_index = bw_measurements['cores'].index(cores)
            bw = bw_measurements['results'][measurement_kernel][run_index]

            # Correct bandwidth due to miss-measurement of write allocation
            # TODO support non-temporal stores and non-write-allocate architectures
            measurement_kernel_info = self.machine['benchmarks']['kernels'][measurement_kernel]
            factor = (float(measurement_kernel_info['read streams']['bytes']) +
                      2.0*float(measurement_kernel_info['write streams']['bytes']) -
                      float(measurement_kernel_info['read+write streams']['bytes'])) / \
                     (float(measurement_kernel_info['read streams']['bytes']) +
                      float(measurement_kernel_info['write streams']['bytes']))
            bw = bw * factor

            performance = arith_intens * float(bw)
            results['mem bottlenecks'].append({
                'performance': PrefixedUnit(performance, 'FLOP/s'),
                'level': (self.machine['memory hierarchy'][cache_level]['level'] + '-' +
                          self.machine['memory hierarchy'][cache_level+1]['level']),
                'arithmetic intensity': arith_intens,
                'bw kernel': measurement_kernel,
                'bandwidth': bw})
            if performance <= results.get('min performance', performance):
                results['bottleneck level'] = len(results['mem bottlenecks'])-1
                results['min performance'] = performance
        return results

    def analyze(self):
        self.results = self.calculate_cache_access()

    def report(self):
        max_flops = self.machine['clock']*sum(self.machine['FLOPs per cycle']['DP'].values())
        max_flops.unit = "FLOP/s"
        if self._args and self._args.verbose >= 1:
            print('Bottlnecks:')
            print('  level | a. intensity |   performance   |  bandwidth | bandwidth kernel')
            print('--------+--------------+-----------------+------------+-----------------')
            print('    CPU |              | {:>15} |            |'.format(max_flops))
            for b in self.results['mem bottlenecks']:
                print('{level:>7} | {arithmetic intensity:>5.2} FLOP/b | {performance:>15} |'
                      ' {bandwidth:>10} | {bw kernel:<8}'.format(**b))
            print()

        # TODO support SP
        if self.results['min performance'] > max_flops:
            # CPU bound
            print('CPU bound')
            print('{!s} due to CPU max. FLOP/s'.format(max_flops))
        else:
            # Cache or mem bound
            print('Cache or mem bound')

            bottleneck = self.results['mem bottlenecks'][self.results['bottleneck level']]
            print('{!s} due to {} transfer bottleneck (bw with from {} benchmark)'.format(
                bottleneck['performance'],
                bottleneck['level'],
                bottleneck['bw kernel']))
            print('Arithmetic Intensity: {:.2f}'.format(bottleneck['arithmetic intensity']))

class RooflineIACA(Roofline):
    """
    class representation of the Roofline Model (with IACA throughput analysis)

    more info to follow...
    """

    name = "Roofline (with IACA throughput)"

    @classmethod
    def configure_arggroup(cls, parser):
        pass

    def __init__(self, kernel, machine, args=None, parser=None):
        """
        *kernel* is a Kernel object
        *machine* describes the machine (cpu, cache and memory) characteristics
        *args* (optional) are the parsed arguments from the comand line
        if *args* is given also *parser* has to be provided
        """
        Roofline.__init__(self, kernel, machine, args, parser)

    def analyze(self):
        self.results = self.calculate_cache_access()
        
        # For the IACA/CPU analysis we need to compile and assemble
        asm_name = self.kernel.compile(compiler_args=self.machine['icc architecture flags'])
        bin_name = self.kernel.assemble(
            asm_name, iaca_markers=True, asm_block=self._args.asm_block)

        iaca_output = subprocess.check_output(
            ['iaca.sh', '-64', '-arch', self.machine['micro-architecture'], bin_name])

        # Get total cycles per loop iteration
        match = re.search(
            r'^Block Throughput: ([0-9\.]+) Cycles', iaca_output, re.MULTILINE)
        assert match, "Could not find Block Throughput in IACA output."
        block_throughput = float(match.groups()[0])

        # Find ports and cyles per port
        ports = filter(lambda l: l.startswith('|  Port  |'), iaca_output.split('\n'))
        cycles = filter(lambda l: l.startswith('| Cycles |'), iaca_output.split('\n'))
        assert ports and cycles, "Could not find ports/cylces lines in IACA output."
        ports = map(str.strip, ports[0].split('|'))[2:]
        cycles = map(str.strip, cycles[0].split('|'))[2:]
        port_cycles = []
        for i in range(len(ports)):
            if '-' in ports[i] and ' ' in cycles[i]:
                subports = map(str.strip, ports[i].split('-'))
                subcycles = filter(bool, cycles[i].split(' '))
                port_cycles.append((subports[0], float(subcycles[0])))
                port_cycles.append((subports[0]+subports[1], float(subcycles[1])))
            elif ports[i] and cycles[i]:
                port_cycles.append((ports[i], float(cycles[i])))
        port_cycles = dict(port_cycles)

        match = re.search(r'^Total Num Of Uops: ([0-9]+)', iaca_output, re.MULTILINE)
        assert match, "Could not find Uops in IACA output."
        uops = float(match.groups()[0])

        # Normalize to cycles per cacheline
        elements_per_block = self.kernel.blocks[self.kernel.block_idx][1]['loop_increment']
        block_size = elements_per_block*8  # TODO support SP
        block_to_cl_ratio = float(self.machine['cacheline size'])/block_size

        port_cycles = dict(map(lambda i: (i[0], i[1]*block_to_cl_ratio), port_cycles.items()))
        uops = uops*block_to_cl_ratio
        cl_throughput = block_throughput*block_to_cl_ratio
        flops_per_element = sum(self.kernel._flops.values())

        # Create result dictionary
        self.results.update({
            'cpu bottleneck': {
                'port cycles': port_cycles,
                'cl throughput': cl_throughput,
                'uops': uops,
                'performance':
                    self.machine['clock']/block_throughput*elements_per_block*flops_per_element}})
        self.results['cpu bottleneck']['performance'].unit = 'FLOP/s'

    def report(self):
        cpu_flops = PrefixedUnit(self.results['cpu bottleneck']['performance'], "FLOP/s")
        if self._args and self._args.verbose >= 1:
            print('Bottlnecks:')
            print('  level | a. intensity |   performance   |  bandwidth | bandwidth kernel')
            print('--------+--------------+-----------------+------------+-----------------')
            print('    CPU |              | {:>15} |            |'.format(cpu_flops))
            for b in self.results['mem bottlenecks']:
                print('{level:>7} | {arithmetic intensity:>5.2} FLOP/b | {performance:>15} |'
                      ' {bandwidth:>10} | {bw kernel:<8}'.format(**b))
            print()
            print('IACA analisys:')
            print(self.results['cpu bottleneck'])

        # TODO support SP
        if float(self.results['min performance']) > float(cpu_flops):
            # CPU bound
            print('CPU bound')
            print('{!s} due to CPU bottleneck'.format(cpu_flops))
        else:
            # Cache or mem bound
            print('Cache or mem bound')

            bottleneck = self.results['mem bottlenecks'][self.results['bottleneck level']]
            print('{!s} due to {} transfer bottleneck (bw with from {} benchmark)'.format(
                bottleneck['performance'],
                bottleneck['level'],
                bottleneck['bw kernel']))
            print('Arithmetic Intensity: {:.2f}'.format(bottleneck['arithmetic intensity']))



#!/usr/bin/env python
# pylint: disable=W0142

from __future__ import print_function
from pprint import pprint
from functools import reduce as reduce_
import operator
import copy
import sys
import subprocess
import re
import imp

try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    plot_support = True
except ImportError:
    plot_support = False

from kerncraft.intervals import Intervals

# Datatype sizes in bytes
datatype_size = {'double': 8, 'float': 4}


def blocking(indices, block_size, initial_boundary=0):
    '''
    splits list of integers into blocks of block_size. returns block indices.

    first block element is located at initial_boundary (default 0).

    >>> blocking([0, -1, -2, -3, -4, -5, -6, -7, -8, -9], 8)
    [0,-1]
    >>> blocking([0], 8)
    [0]
    '''
    blocks = []

    for idx in indices:
        bl_idx = (idx-initial_boundary)/block_size
        if bl_idx not in blocks:
            blocks.append(bl_idx)
    blocks.sort()

    return blocks


def flatten_dict(d):
    '''
    transforms 2d-dict d[i][k] into a new 1d-dict e[(i,k)] with 2-tuple keys
    '''
    e = {}
    for k in d.keys():
        for l in d[k].keys():
            e[(k, l)] = d[k][l]
    return e


class ECMData:
    """
    class representation of the Execution-Cache-Memory Model (only the data part)

    more info to follow...
    """

    name = "Execution-Cache-Memory (data transfers only)"

    @classmethod
    def configure_arggroup(cls, parser):
        pass

    def __init__(self, kernel, machine, args=None, parser=None):
        """
        *kernel* is a Kernel object
        *machine* describes the machine (cpu, cache and memory) characteristics
        *args* (optional) are the parsed arguments from the comand line
        """
        self.kernel = kernel
        self.machine = machine
        self._args = args
        self._parser = parser

        if args:
            # handle CLI info
            pass

    def _calculate_relative_offset(self, name, access_dimensions):
        '''
        returns the offset from the iteration center in number of elements and the order of indices
        used in access.
        '''
        offset = 0
        base_dims = self.kernel._variables[name][1]

        for dim, offset_info in enumerate(access_dimensions):
            offset_type, idx_name, dim_offset = offset_info
            assert offset_type == 'rel', 'Only relative access to arrays is supported at the moment'

            if offset_type == 'rel':
                offset += dim_offset*reduce_(operator.mul, base_dims[dim+1:], 1)
            else:
                # should not happen
                pass

        return offset

    def _calculate_iteration_offset(self, name, index_order, loop_index):
        '''
        returns the offset from one to the next iteration using *loop_index*.
        *index_order* is the order used by the access dimensions e.g. 'ijk' corresponse to [i][j][k]
        *loop_index* specifies the loop to be used for iterations (this is typically the inner
        moste one)
        '''
        offset = 0
        base_dims = self.kernel._variables[name][1]

        for dim, index_name in enumerate(index_order):
            if loop_index == index_name:
                offset += reduce_(operator.mul, base_dims[dim+1:], 1)

        return offset

    def _get_index_order(self, access_dimensions):
        '''Returns the order of indices used in *access_dimensions*.'''
        return ''.join(map(lambda d: d[1], access_dimensions))

    def _expand_to_cacheline_blocks(self, first, last):
        '''
        Returns first and last values wich align with cacheline blocks, by increasing range.
        '''
        # TODO how to handle multiple datatypes (with different size)?
        element_size = datatype_size['double']
        elements_per_cacheline = int(float(self.machine['cacheline size'])) / element_size

        first = first - first % elements_per_cacheline
        last = last - last % elements_per_cacheline + elements_per_cacheline - 1

        return [first, last]

    def calculate_cache_access(self):
        results = {}

        read_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}
        write_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}
        iteration_offsets = {var_name: dict() for var_name in self.kernel._variables.keys()}

        # TODO how to handle multiple datatypes (with different size)?
        element_size = datatype_size['double']
        elements_per_cacheline = int(float(self.machine['cacheline size'])) / element_size

        loop_order = ''.join(map(lambda l: l[0], self.kernel._loop_stack))

        for var_name in self.kernel._variables.keys():
            var_type, var_dims = self.kernel._variables[var_name]

            # Skip the following access: (they are hopefully kept in registers)
            #   - scalar values
            if var_dims is None:
                continue
            #   - access does not change with inner-most loop index
            writes = filter(
                lambda acs: loop_order[-1] in map(lambda a: a[1], acs),
                self.kernel._destinations.get(var_name, []))
            reads = filter(
                lambda acs: loop_order[-1] in map(lambda a: a[1], acs),
                self.kernel._sources.get(var_name, []))

            # Compile access pattern
            for r in reads:
                offset = self._calculate_relative_offset(var_name, r)
                idx_order = self._get_index_order(r)
                read_offsets[var_name].setdefault(idx_order, [])
                read_offsets[var_name][idx_order].append(offset)
            for w in writes:
                offset = self._calculate_relative_offset(var_name, w)
                idx_order = self._get_index_order(w)
                write_offsets[var_name].setdefault(idx_order, [])
                write_offsets[var_name][idx_order].append(offset)

            # Do unrolling so that one iteration equals one cacheline worth of workload:
            # unrolling is done on inner-most loop only!
            for i in range(1, elements_per_cacheline):
                for r in reads:
                    idx_order = self._get_index_order(r)
                    offset = self._calculate_relative_offset(var_name, r)
                    offset += i * self._calculate_iteration_offset(
                        var_name, idx_order, loop_order[-1])
                    read_offsets[var_name][idx_order].append(offset)

                    # Remove multiple access to same offsets
                    read_offsets[var_name][idx_order] = \
                        sorted(list(set(read_offsets[var_name][idx_order])), reverse=True)

                for w in writes:
                    idx_order = self._get_index_order(w)
                    offset = self._calculate_relative_offset(var_name, w)
                    offset += i * self._calculate_iteration_offset(
                        var_name, idx_order, loop_order[-1])
                    write_offsets[var_name][idx_order].append(offset)

                    # Remove multiple access to same offsets
                    write_offsets[var_name][idx_order] = \
                        sorted(list(set(write_offsets[var_name][idx_order])), reverse=True)

        # initialize misses and hits
        misses = {}
        hits = {}
        evicts = {}
        total_misses = {}
        total_hits = {}
        total_evicts = {}
        total_lines_misses = {}
        total_lines_hits = {}
        total_lines_evicts = {}

        self.results = {'memory hierarchy': [], 'cycles': []}

        # Check for layer condition towards all cache levels (except main memory/last level)
        for i, cache_info in list(enumerate(self.machine['memory hierarchy']))[:-1]:
            cache_size = int(float(cache_info['size per group']))
            cache_cycles = cache_info['cycles per cacheline transfer']

            trace_length = 0
            updated_length = True
            while updated_length:
                updated_length = False

                # Initialize cache, misses, hits and evicts for current level
                cache = {}
                misses[i] = {}
                hits[i] = {}
                evicts[i] = {}

                # We consider everythin a miss in the beginning
                # TODO here read and writes are treated the same, this implies write-allocate
                #      to support nontemporal stores, this needs to be changed
                for name in read_offsets.keys()+write_offsets.keys():
                    cache[name] = {}
                    misses[i][name] = {}
                    hits[i][name] = {}

                    for idx_order in read_offsets[name].keys()+write_offsets[name].keys():
                        cache[name][idx_order] = Intervals()
                        if i-1 not in misses:
                            misses[i][name][idx_order] = sorted(
                                read_offsets.get(name, {}).get(idx_order, []) +
                                write_offsets.get(name, {}).get(idx_order, []),
                                reverse=True)
                        else:
                            misses[i][name][idx_order] = list(misses[i-1][name][idx_order])
                        hits[i][name][idx_order] = []

                # Caches are still empty (thus only misses)
                trace_count = 0
                cache_used_size = 0

                # Now we trace the cache access backwards (in time/iterations) and check for hits
                for var_name in misses[i].keys():
                    for idx_order in misses[i][var_name].keys():
                        iter_offset = self._calculate_iteration_offset(
                            var_name, idx_order, loop_order[-1])

                        # Add cache trace
                        for offset in list(misses[i][var_name][idx_order]):
                            # If already present in cache add to hits
                            if offset in cache[var_name][idx_order]:
                                misses[i][var_name][idx_order].remove(offset)

                                # We might have multiple hits on the same offset (e.g in DAXPY)
                                if offset not in hits[i][var_name][idx_order]:
                                    hits[i][var_name][idx_order].append(offset)

                            # Add cache, we can do this since misses are sorted in reverse order of
                            # access and we assume LRU cache replacement policy
                            if iter_offset <= elements_per_cacheline:
                                # iterations overlap, thus we can savely add the whole range
                                cached_first, cached_last = self._expand_to_cacheline_blocks(
                                    offset-iter_offset*trace_length, offset+1)
                                cache[var_name][idx_order] &= Intervals(
                                    [cached_first, cached_last+1], sane=True)
                            else:
                                # There is no overlap, we can append the ranges onto one another
                                # TODO optimize this code section (and maybe merge with above)
                                new_cache = [self._expand_to_cacheline_blocks(o, o) for o in range(
                                    offset-iter_offset*trace_length, offset+1, iter_offset)]
                                new_cache = Intervals(*new_cache, sane=True)
                                cache[var_name][idx_order] &= new_cache

                        trace_count += len(cache[var_name][idx_order]._data)
                        cache_used_size += len(cache[var_name][idx_order])*element_size

                # Calculate new possible trace_length according to free space in cache
                # TODO take CL blocked access into account
                # TODO make /2 customizable
                new_trace_length = trace_length + \
                    ((cache_size/2 - cache_used_size)/trace_count)/element_size

                if new_trace_length > trace_length:
                    trace_length = new_trace_length
                    updated_length = True

                # All writes to require the data to be evicted eventually
                evicts[i] = {var_name: dict() for var_name in self.kernel._variables.keys()}
                for name in write_offsets.keys():
                    for idx_order in write_offsets[name].keys():
                        evicts[i][name][idx_order] = list(write_offsets[name][idx_order])

            # Compiling stats
            total_misses[i] = sum(map(lambda l: sum(map(len, l.values())), misses[i].values()))
            total_hits[i] = sum(map(lambda l: sum(map(len, l.values())), hits[i].values()))
            total_evicts[i] = sum(map(lambda l: sum(map(len, l.values())), evicts[i].values()))

            total_lines_misses[i] = sum(map(
                lambda o: sum(map(lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                misses[i].values()))
            total_lines_hits[i] = sum(map(lambda o: sum(map(
                lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                hits[i].values()))
            total_lines_evicts[i] = sum(map(lambda o: sum(map(
                lambda n: len(blocking(n, elements_per_cacheline)), o.values())),
                evicts[i].values()))

            if cache_cycles:
                # Cache
                cycles = (total_lines_misses[i] + total_lines_evicts[i]) * \
                    cache_cycles
            else:
                # Memory
                cycles = (total_lines_misses[i] + total_lines_evicts[i]) * \
                    elements_per_cacheline * element_size * \
                    float(self.machine['clock']) / \
                    float(self.machine['memory hierarchy'][-2]['bandwidth'])

            self.results['memory hierarchy'].append({
                'index': i,
                'level': '{}'.format(cache_info['level']),
                'total misses': total_misses[i],
                'total hits': total_hits[i],
                'total evicts': total_evicts[i],
                'total lines misses': total_lines_misses[i],
                'total lines hits': total_lines_hits[i],
                'total lines evicts': total_lines_evicts[i],
                'trace length': trace_length,
                'misses': misses[i],
                'hits': hits[i],
                'evicts': evicts[i],
                'cycles': cycles})
            self.results['cycles'].append((
                '{}-{}'.format(cache_info['level'], self.machine['memory hierarchy'][i+1]['level']),
                cycles))

            # TODO remove the following by makeing testcases more versatile:
            self.results[
                '{}-{}'.format(cache_info['level'], self.machine['memory hierarchy'][i+1]['level'])
                ] = cycles

        return results

    def analyze(self):
        self._results = self.calculate_cache_access()

    def report(self):
        if self._args and self._args.verbose > 0:
            for r in self.results['memory hierarchy']:
                print('Trace legth per access in {}:'.format(r['level']), r['trace length'])
                print('Hits in {}:'.format(r['level']), r['total hits'], r['hits'])
                print('Misses in {}: {} ({}CL):'.format(
                    r['level'], r['total misses'], r['total lines misses']),
                    r['misses'])
                print('Evicts from {} {} ({}CL):'.format(
                    r['level'], r['total evicts'], r['total lines evicts']),
                    r['evicts'])

        for level, cycles in self.results['cycles']:
            print('{} = {}cy'.format(level, cycles))


class ECMCPU:
    """
    class representation of the Execution-Cache-Memory Model (only the operation part)

    more info to follow...
    """

    name = "Execution-Cache-Memory (CPU operations only)"

    @classmethod
    def configure_arggroup(cls, parser):
        pass

    def __init__(self, kernel, machine, args=None, parser=None):
        """
        *kernel* is a Kernel object
        *machine* describes the machine (cpu, cache and memory) characteristics
        *args* (optional) are the parsed arguments from the comand line
        if *args* is given also *parser* has to be provided
        """
        self.kernel = kernel
        self.machine = machine
        self._args = args
        self._parser = parser

        if args:
            # handle CLI info
            if self._args.asm_block not in ['auto', 'manual']:
                try:
                    self._args.asm_block = int(args.asm_block)
                except ValueError:
                    parser.error('--asm-block can only be "auto", "manual" or an integer')

    def analyze(self):
        # For the IACA/CPU analysis we need to compile and assemble
        asm_name = self.kernel.compile(compiler_args=self.machine['icc architecture flags'])
        bin_name = self.kernel.assemble(
            asm_name, iaca_markers=True, asm_block=self._args.asm_block)

        iaca_output = subprocess.check_output(
            ['iaca.sh', '-64', '-arch', self.machine['micro-architecture'], bin_name])

        # Get total cycles per loop iteration
        match = re.search(
            r'^Block Throughput: ([0-9\.]+) Cycles', iaca_output, re.MULTILINE)
        assert match, "Could not find Block Throughput in IACA output."
        block_throughput = float(match.groups()[0])

        # Find ports and cyles per port
        ports = filter(lambda l: l.startswith('|  Port  |'), iaca_output.split('\n'))
        cycles = filter(lambda l: l.startswith('| Cycles |'), iaca_output.split('\n'))
        assert ports and cycles, "Could not find ports/cylces lines in IACA output."
        ports = map(str.strip, ports[0].split('|'))[2:]
        cycles = map(str.strip, cycles[0].split('|'))[2:]
        port_cycles = []
        for i in range(len(ports)):
            if '-' in ports[i] and ' ' in cycles[i]:
                subports = map(str.strip, ports[i].split('-'))
                subcycles = filter(bool, cycles[i].split(' '))
                port_cycles.append((subports[0], float(subcycles[0])))
                port_cycles.append((subports[0]+subports[1], float(subcycles[1])))
            elif ports[i] and cycles[i]:
                port_cycles.append((ports[i], float(cycles[i])))
        port_cycles = dict(port_cycles)

        match = re.search(r'^Total Num Of Uops: ([0-9]+)', iaca_output, re.MULTILINE)
        assert match, "Could not find Uops in IACA output."
        uops = float(match.groups()[0])

        # Normalize to cycles per cacheline
        block_elements = self.kernel.blocks[self.kernel.block_idx][1]['loop_increment']
        block_size = block_elements*8  # TODO support SP
        block_to_cl_ratio = float(self.machine['cacheline size'])/block_size

        port_cycles = dict(map(lambda i: (i[0], i[1]*block_to_cl_ratio), port_cycles.items()))
        uops = uops*block_to_cl_ratio
        block_throughput = block_throughput*block_to_cl_ratio

        # Compile most relevant information
        T_OL = max([v for k, v in port_cycles.items() if k in self.machine['overlapping ports']])
        T_nOL = max(
            [v for k, v in port_cycles.items() if k in self.machine['non-overlapping ports']])

        # Create result dictionary
        self.results = {
            'port cycles': port_cycles,
            'block throughput': block_throughput,
            'uops': uops,
            'T_nOL': T_nOL,
            'T_OL': T_OL}

    def report(self):
        if self._args and self._args.verbose > 0:
            print('Ports and cycles:', self.results['port cycles'])
            print('Uops:', self.results['uops'])

            print('Throughput: {}cy per CL'.format(self.results['block throughput']))

        print('T_nOL = {}cy'.format(self.results['T_nOL']))
        print('T_OL = {}cy'.format(self.results['T_OL']))


class ECM:
    """
    class representation of the Execution-Cache-Memory Model (data and operations)

    more info to follow...
    """

    name = "Execution-Cache-Memory"

    @classmethod
    def configure_arggroup(cls, parser):
        # they are being configured in ECMData and ECMCPU
        parser.add_argument(
            '--ecm-plot',
            help='Filename to save ECM plot to (supported extensions: pdf, png, svg and eps)')

    def __init__(self, kernel, machine, args=None, parser=None):
        """
        *kernel* is a Kernel object
        *machine* describes the machine (cpu, cache and memory) characteristics
        *args* (optional) are the parsed arguments from the comand line
        """
        self.kernel = kernel
        self.machine = machine
        self._args = args

        if args:
            # handle CLI info
            pass

        self._CPU = ECMCPU(kernel, machine, args, parser)
        self._data = ECMData(kernel, machine, args, parser)

    def analyze(self):
        self._CPU.analyze()
        self._data.analyze()
        self.results = copy.deepcopy(self._CPU.results)
        self.results.update(copy.deepcopy(self._data.results))

    def report(self):
        report = ''
        if self._args and self._args.verbose > 0:
            self._CPU.report()
            self._data.report()

        report += '{{ {} || {} | {} }}'.format(
            self.results['T_OL'],
            self.results['T_nOL'],
            ' | '.join([str(i[1]) for i in self.results['cycles']]))

        print(report)

        if self._args and self._args.ecm_plot:
            assert plot_support, "matplotlib couldn't be imported. Plotting is not supported."

            fig = plt.figure(frameon=False)
            ax = fig.add_subplot(1, 1, 1)

            sorted_overlapping_ports = sorted(
                map(lambda p: (p, self.results['port cycles'][p]),
                    self.machine['overlapping ports']),
                key=lambda x: x[1])

            yticks_labels = []
            yticks = []
            xticks_labels = []
            xticks = []

            # Plot configuration
            height = 0.8

            i = 0
            # T_OL
            colors = [(254./255, 177./255., 178./255.)] + [(255./255., 255./255., 255./255.)] * \
                (len(sorted_overlapping_ports) - 1)
            for p, c in sorted_overlapping_ports:
                ax.barh(i, c, height, align='center', color=colors.pop())
                yticks_labels.append(p)
                yticks.append(i)
                i += 1
            xticks.append(sorted_overlapping_ports[-1][1])
            xticks_labels.append('{:.1f}'.format(sorted_overlapping_ports[-1][1]))

            # T_nOL + memory transfers
            y = 0
            colors = [(187./255., 255/255., 188./255.)] * (len(self.results['cycles'])) + \
                [(119./255, 194./255., 255./255.)]
            for k, v in [('T_nOL', self.results['T_nOL'])]+self.results['cycles']:
                ax.barh(i, v, height, y, align='center', color=colors.pop())
                ax.text(y+v/2.0, i, k, ha='center', va='center')
                xticks.append(y+v)
                xticks_labels.append('{:.1f}'.format(y+v))
                y += v
            yticks_labels.append('LD')
            yticks.append(i)

            ax.tick_params(axis='y', which='both', left='off', right='off')
            ax.tick_params(axis='x', which='both', top='off')
            ax.set_xlabel('t [cy]')
            ax.set_ylabel('execution port')
            ax.set_yticks(yticks)
            ax.set_yticklabels(yticks_labels)
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticks_labels, rotation='vertical')
            ax.xaxis.grid(alpha=0.7, linestyle='--')
            fig.savefig(self._args.ecm_plot)



#!/usr/bin/env python

from __future__ import print_function

import argparse
import ast
import sys
import os.path

import models
from kernel import Kernel
from machinemodel import MachineModel


class AppendStringInteger(argparse.Action):
    """Action to append a string and integer"""
    def __call__(self, parser, namespace, values, option_string=None):
        message = ''
        if len(values) != 2:
            message = 'requires 2 arguments'
        else:
            try:
                values[1] = int(values[1])
            except ValueError:
                message = ('second argument requires an integer')

        if message:
            raise argparse.ArgumentError(self, message)

        if hasattr(namespace, self.dest):
            getattr(namespace, self.dest).append(values)
        else:
            setattr(namespace, self.dest, [values])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--machine', '-m', type=file, required=True,
                        help='Path to machine description yaml file.')
    parser.add_argument('--pmodel', '-p', choices=models.__all__, required=True, action='append',
                        default=[], help='Performance model to apply')
    parser.add_argument('-D', '--define', nargs=2, metavar=('KEY', 'VALUE'), default=[],
                        action=AppendStringInteger,
                        help='Define constant to be used in C code. Values must be integers. '
                             'Overwrites constants from testcase file.')
    parser.add_argument('--testcases', '-t', action='store_true',
                        help='Use testcases file.')
    parser.add_argument('--testcase-index', '-i', metavar='INDEX', type=int, default=0,
                        help='Index of testcase in testcase file. If not given, all cases are '
                             'executed.')
    parser.add_argument('--verbose', '-v', action='count',
                        help='Increases verbosity level.')
    parser.add_argument('code_file', metavar='FILE', type=argparse.FileType(), nargs='+',
                        help='File with loop kernel C code')
    parser.add_argument('--asm-block', metavar='BLOCK', default='auto',
                        help='Number of ASM block to mark for IACA, "auto" for automatic '
                             'selection or "manual" for interactiv selection.')
    for m in models.__all__:
        ag = parser.add_argument_group('arguments for '+m+' model', getattr(models, m).name)
        getattr(models, m).configure_arggroup(ag)

    # BUSINESS LOGIC IS FOLLOWING
    args = parser.parse_args()
    
    # Checking arguments
    if args.asm_block not in ['auto', 'manual']:
        try:
            args.asm_block = int(args.asm_block)
        except ValueError:
            parser.error('--asm-block can only be "auto", "manual" or an integer')

    # machine information
    # Read machine description
    machine = MachineModel(args.machine.name)

    # process kernels and testcases
    for code_file in args.code_file:
        code = code_file.read()

        # Add constants from testcase file
        if args.testcases:
            testcases_file = open(os.path.splitext(code_file.name)[0]+'.testcases')
            testcases = ast.literal_eval(testcases_file.read())
            if args.testcase_index:
                testcases = [testcases[args.testcase_index]]
        else:
            testcases = [{'constants': {}}]

        for testcase in testcases:

            print('='*80 + '\n{:^80}\n'.format(code_file.name) + '='*80)

            kernel = Kernel(code, filename=code_file.name)

            assert 'constants' in testcase, "Could not find key 'constants' in testcase file."
            for k, v in testcase['constants']:
                kernel.set_constant(k, v)

            # Add constants from define arguments
            for k, v in args.define:
                kernel.set_constant(k, v)

            kernel.process()
            kernel.print_kernel_code()
            print()
            kernel.print_variables_info()
            kernel.print_constants_info()
            kernel.print_kernel_info()

            for model_name in set(args.pmodel):
                model = getattr(models, model_name)(kernel, machine, args, parser)

                model.analyze()
                model.report()

                # TODO take care of different performance models
                if 'results-to-compare' in testcase:
                    failed = False
                    for key, value in model.results.items():
                        if key in testcase['results-to-compare']:
                            correct_value = float(testcase['results-to-compare'][key])
                            diff = abs(value - correct_value)
                            if diff > correct_value*0.1:
                                print("Test values did not match: {} should have been {}, but was "
                                      "{}.".format(key, correct_value, value))
                                failed = True
                            elif diff:
                                print("Small difference from theoretical value: {} should have "
                                      "been {}, but was {}.".format(key, correct_value, value))
                    if failed:
                        sys.exit(1)

if __name__ == '__main__':
    main()
